/*
 * Noise generation for PuTTY's cryptographic random number
 * generator. (based on windows/winnoise.c)
 */

#include <basic.h>
#include <bstdio.h>
#include <btron/proctask.h>
#include <btron/file.h>
#include <btron/memory.h>
#include <btron/clk.h>
#include <btron/hmi.h>
#include <btron/event.h>

#include "putty.h"
#include "unix.h"
#include "ssh.h"
#include "storage.h"
#include "tcsliteral.h"

/*
 * This function is called once, at PuTTY startup, and will do some
 * seriously silly things like listing directories and getting disk
 * free space and a process snapshot.
 */
void noise_get_heavy(void (*func) (void *, int))
{
  TC sys_path[] = {TCS_SYS_PATH};
  LINK lnk;
  W fd, pid, err;
  int first = TRUE;

  get_lnk(sys_path, &lnk, F_NORM);
  fd = opn_fil(&lnk, F_READ, NULL);
  if (fd >= 0) {
    while (fnd_lnk(fd, (first ? F_TOPEND : F_NFWD), &lnk, 0, NULL) >= 0) {
      first = FALSE;
      err = rea_rec(fd, 0, (B*)&lnk, sizeof(lnk), NULL, NULL);
      if (err >= 0) {
        func(&lnk, sizeof(lnk));
      }
    }
  }

  pid = prc_sts(0, NULL, NULL);
  func(&pid, sizeof(pid));

  read_random_seed(func);
  /* Update the seed immediately, in case another instance uses it. */
  random_save_seed();
}

void random_save_seed(void)
{
  int len;
  void *data;

  if (random_active) {
    random_get_savedata(&data, &len);
    write_random_seed(data, len);
    sfree(data);
  }
}

/*
 * This function is called every time the random pool needs
 * stirring, and will acquire the system time in all available
 * forms.
 */
void noise_get_light(void (*func) (void *, int))
{
  STIME systime;
  long ticks;

  if (get_tim(&systime, NULL) >= 0) {
    func(&systime, sizeof(systime));
  }

  ticks = GETTICKCOUNT();
  random_add_noise(&ticks, sizeof(ticks));
}

/*
 * This function is called on a timer, and it will monitor
 * frequently changing quantities such as the state of physical and
 * virtual memory, the state of the process's message queue, which
 * window is in the foreground, which owns the clipboard, etc.
 */
void noise_regular(void)
{
  W w;
  PNT pt;
  P_INFO pinf;
  M_STATE memstat;

  w = wget_act(NULL);
  random_add_noise(&w, sizeof(w));
#if 0 /* BTRON */
  w = GetCapture();
  random_add_noise(&w, sizeof(w));
  w = GetClipboardOwner();
  random_add_noise(&w, sizeof(w));
  z = GetQueueStatus(QS_ALLEVENTS);
  random_add_noise(&z, sizeof(z));
#endif

  get_pdp(&pt);
  random_add_noise(&pt, sizeof(pt));

  mbk_sts(&memstat);
  random_add_noise(&memstat, sizeof(memstat));

  get_inf(0, &pinf);
  random_add_noise(&pinf, sizeof(pinf));
}

/*
 * This function is called on every keypress or mouse move, and
 * will add the current Windows time and performance monitor
 * counter to the noise pool. It gets the scan code or mouse
 * position passed in.
 */
void noise_ultralight(unsigned long data)
{
  long ticks;
  P_INFO pinf;

  random_add_noise(&data, sizeof(data));

  ticks = GETTICKCOUNT();
  random_add_noise(&ticks, sizeof(ticks));

  if (get_inf(0, &pinf) >= 0) {
    UW perftime = pinf.utime + pinf.stime;
    random_add_noise(&perftime, sizeof(perftime));
  }
}
