/*
 * uxsel.c
 * 
 * This module is a sort of all-purpose interchange for file
 * descriptors. At one end it talks to uxnet.c and pty.c and
 * anything else which might have one or more fds that need
 * select()-type things doing to them during an extended program
 * run; at the other end it talks to pterm.c or uxplink.c or
 * anything else which might have its own means of actually doing
 * those select()-type things.
 */

#include <basic.h>
#include <bassert.h>
#include <btron/bsocket.h>
#include <btron/message.h>
#include <btron/proctask.h>

#include "putty.h"
#include "tree234.h"

//static tree234 *fds;
tree234 *fds;

static void selecter_task(struct fd *fd)
{
  W sel;
  fd_set rfds, wfds, efds;
  int i;
  W err;
  struct fd_input_message msg;

  for (;;) {
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);
    if ((fd->rwx & 4) != 0) {
      FD_SET(fd->fd, &efds);
    }
    if ((fd->rwx & 2) != 0) {
      FD_SET(fd->fd, &wfds);
    }
    if ((fd->rwx & 1) != 0) {
      FD_SET(fd->fd, &rfds);
    }

    sel = so_select(fd->fd + 1, &rfds, &wfds, &efds, NULL);
    if (sel > 0) {
      msg.msg_type = MS_TYPE7;
      msg.msg_size = sizeof(struct fd_input_message) - sizeof(W)*2;
      msg.fd = fd;
      msg.rwx = 
          (FD_ISSET(fd->fd, &efds) ? 4 : 0) 
        | (FD_ISSET(fd->fd, &wfds) ? 2 : 0)
        | (FD_ISSET(fd->fd, &rfds) ? 1 : 0);
      err = snd_msg(0, (MESSAGE*)&msg, WAIT);

      /* wait for completion of message processing by receiver */
      slp_tsk(-1);
    }
    else if (sel < 0) {
      break;
    }
  }

  /* not reached */
  ext_tsk();
}

static int uxsel_fd_cmp(void *av, void *bv)
{
    struct fd *a = (struct fd *)av;
    struct fd *b = (struct fd *)bv;
    if (a->fd < b->fd)
	return -1;
    if (a->fd > b->fd)
	return +1;
    return 0;
}
static int uxsel_fd_findcmp(void *av, void *bv)
{
    int *a = (int *)av;
    struct fd *b = (struct fd *)bv;
    if (*a < b->fd)
	return -1;
    if (*a > b->fd)
	return +1;
    return 0;
}

void uxsel_init(void)
{
    fds = newtree234(uxsel_fd_cmp);
}

/*
 * Here is the interface to fd-supplying modules. They supply an
 * fd, a set of read/write/execute states, and a callback function
 * for when the fd satisfies one of those states. Repeated calls to
 * uxsel_set on the same fd are perfectly legal and serve to change
 * the rwx state (typically you only want to select an fd for
 * writing when you actually have pending data you want to write to
 * it!).
 */

void uxsel_set(int fd, int rwx, uxsel_callback_fn callback)
{
    struct fd *newfd = snew(struct fd);
    struct fd *oldfd;
    P_STATE pst;

    newfd->fd = fd;
    newfd->rwx = rwx;
    newfd->callback = callback;

    oldfd = find234(fds, newfd, NULL);

    /*
     * Because uxsel_set is heavy for BTRON, because it is implemented
     * using `tasks'. Therefore, here we add the check to avoid 
     * unnecessary ter_tsk/cre_tsks to avoid continuous ter_tsk/cre_tsks
     * in the call sequence
     *   net_select_result -> try_send -> uxsel_tell -> uxsel_set
     *   (where uxsel_tell registers net_select_result as callback) 
     */
    if (oldfd != NULL && fd == oldfd->fd && rwx == oldfd->rwx 
        && callback == oldfd->callback) {
      sfree(newfd);
      return;
    }

    if (oldfd) {
//	uxsel_input_remove(oldfd->id);
	ter_tsk(oldfd->tskid);
	del234(fds, oldfd);
	sfree(oldfd);
    }

    add234(fds, newfd);

//  newfd->id = uxsel_input_add(fd, rwx);
    prc_sts(0, &pst, NULL);
    newfd->tskid = cre_tsk((FP)selecter_task, pst.priority, newfd);
}

void uxsel_del(int fd)
{
    struct fd *oldfd = find234(fds, &fd, uxsel_fd_findcmp);
    if (oldfd) {
//	uxsel_input_remove(oldfd->id);
	ter_tsk(oldfd->tskid);
	sfree(oldfd);
    }
}

/*
 * And here is the interface to select-functionality-supplying
 * modules. 
 */

int next_fd(int *state, int *rwx)
{
    struct fd *fd;
    fd = index234(fds, (*state)++);
    if (fd) {
	*rwx = fd->rwx;
	return fd->fd;
    } else
	return -1;
}

int first_fd(int *state, int *rwx)
{
    *state = 0;
    return next_fd(state, rwx);
}

int select_result(int fd, int event)
{
    struct fd *fdstruct = find234(fds, &fd, uxsel_fd_findcmp);
    /*
     * Apparently this can sometimes be NULL. Can't see how, but I
     * assume it means I need to ignore the event since it's on an
     * fd I've stopped being interested in. Sigh.
     */
    if (fdstruct)
        return fdstruct->callback(fd, event);
    else
        return 1;
}
