/*
 * btstore.c: BTRON-specific implementation of the interface defined
 * in storage.h.
 */

#include <basic.h>
#include <bstdio.h>
#include <bstdlib.h>
#include <bstring.h>
#include <bassert.h>
#include <berrno.h>
#include <bctype.h>
#include <limits.h>
#include <tstring.h>
typedef UW COLOR;
#include <tad.h>
#include "putty.h"
#include "storage.h"
#include "tree234.h"
#include "tcsliteral.h"
#include "fileio.h"

#ifdef PATH_MAX
#define FNLEN PATH_MAX
#else
#define FNLEN 1024 /* XXX */
#endif

enum {
    INDEX_DIR, INDEX_HOSTKEYS, INDEX_HOSTKEYS_TMP, INDEX_RANDSEED,
    INDEX_SESSIONDIR, INDEX_SESSION,
};

static const char hex[16] = "0123456789ABCDEF";

static char *mungestr(const char *in)
{
    char *out, *ret;

    if (!in || !*in)
        in = "Default Settings";

    ret = out = snewn(3*strlen(in)+1, char);

    while (*in) {
        /*
         * There are remarkably few punctuation characters that
         * aren't shell-special in some way or likely to be used as
         * separators in some file format or another! Hence we use
         * opt-in for safe characters rather than opt-out for
         * specific unsafe ones...
         */
	if (*in!='+' && *in!='-' && *in!='.' && *in!='@' && *in!='_' &&
            !(*in >= '0' && *in <= '9') &&
            !(*in >= 'A' && *in <= 'Z') &&
            !(*in >= 'a' && *in <= 'z')) {
	    *out++ = '%';
	    *out++ = hex[((unsigned char) *in) >> 4];
	    *out++ = hex[((unsigned char) *in) & 15];
	} else
	    *out++ = *in;
	in++;
    }
    *out = '\0';
    return ret;
}

static char *unmungestr(const char *in)
{
    char *out, *ret;
    out = ret = snewn(strlen(in)+1, char);
    while (*in) {
	if (*in == '%' && in[1] && in[2]) {
	    int i, j;

	    i = in[1] - '0';
	    i -= (i > 9 ? 7 : 0);
	    j = in[2] - '0';
	    j -= (j > 9 ? 7 : 0);

	    *out++ = (i << 4) + j;
	    in += 3;
	} else {
	    *out++ = *in++;
	}
    }
    *out = '\0';
    return ret;
}

/*
 * Reading settings, for the moment, is done by retrieving X
 * resources from the X display. When we introduce disk files, I
 * think what will happen is that the X resources will override
 * PuTTY's inbuilt defaults, but that the disk files will then
 * override those. This isn't optimal, but it's the best I can
 * immediately work out.
 * FIXME: the above comment is a bit out of date. Did it happen?
 */

struct keyval {
    const char *key;
    const char *value;
};

int keycmp(void *av, void *bv)
{
    struct keyval *a = (struct keyval *)av;
    struct keyval *b = (struct keyval *)bv;
    return strcmp(a->key, b->key);
}

/*
 * Read an entire line of text from a file. Return a buffer
 * malloced to be as big as necessary (caller must free).
 */
TC *fgetline_tc(FILE *fp)
{
    TC *ret = snewn(512, TC);
    int size = 512, len = 0;
    while (fgets_tc(ret + len, size - len, fp)) {
	len += tc_strlen(ret + len);
	if (ret[len-1] == TK_NL)
	    break;		       /* got a newline, we're done */
	size = len + 512;
	ret = sresize(ret, size, TC);
    }
    if (len == 0) {		       /* first fgets returned NULL */
	sfree(ret);
	return NULL;
    }
    ret[len] = TNULL;
    return ret;
}

void *open_settings_r(LINK *link)
{
  FILE *fp;
  struct Filename fn;
  TC *line_tc;
  char *line;
  tree234 *ret;

  fn.tag = FILENAME_LINK;
  fn.link = *link;

  fp = f_open_rec(fn, "r", 0, RT_TADDATA);
  if (!fp)
    return NULL;		       /* can't open */

  ret = newtree234(keycmp);

  while ( (line_tc = fgetline_tc(fp)) ) {
    char *value;
    struct keyval *kv;
    line = snewn(tc_strlen(line_tc)*2+1, char);
    tcstoeucs(line, line_tc);
    value = strchr(line, '=');

    if (!value || line[0] == '#') {
      sfree(line);
      sfree(line_tc);
      continue;
    }
    *value++ = '\0';
    value[strcspn(value, "\r\n")] = '\0';   /* trim trailing NL */

    kv = snew(struct keyval);
    kv->key = dupstr(line);
    kv->value = dupstr(value);
    add234(ret, kv);

    sfree(line);
    sfree(line_tc);
  }

  fclose(fp);

  return ret;
}

char *read_setting_s(void *handle, const char *key, char *buffer, int buflen)
{
    tree234 *tree = (tree234 *)handle;
    const char *val = NULL;
    struct keyval tmp, *kv;

    tmp.key = key;
    if (tree != NULL &&
        (kv = find234(tree, &tmp, NULL)) != NULL) {
        val = kv->value;
        assert(val != NULL);
    } 

    if (!val)
	return NULL;
    else {
	strncpy(buffer, val, buflen);
	buffer[buflen-1] = '\0';
	return buffer;
    }
}

int read_setting_i(void *handle, const char *key, int defvalue)
{
    tree234 *tree = (tree234 *)handle;
    const char *val = NULL;
    struct keyval tmp, *kv;

    tmp.key = key;
    if (tree != NULL &&
        (kv = find234(tree, &tmp, NULL)) != NULL) {
        val = kv->value;
        assert(val != NULL);
    } 

    if (!val)
	return defvalue;
    else
	return atoi(val);
}

void close_settings_r(void *handle)
{
    tree234 *tree = (tree234 *)handle;
    struct keyval *kv;

    if (!tree)
        return;

    while ( (kv = index234(tree, 0)) != NULL) {
        del234(tree, kv);
        sfree((char *)kv->key);
        sfree((char *)kv->value);
        sfree(kv);
    }

    freetree234(tree);
}

#if 0 /* BTRON TODO */
int read_setting_fontspec(void *handle, const char *name, FontSpec *result)
{
    return !!read_setting_s(handle, name, result->name, sizeof(result->name));
}
#endif

int read_setting_filename(void *handle, const char *name, Filename *result)
{
  int i, v;
  char buf[512];
  B* target;

  if (read_setting_s(handle, name, buf, sizeof(buf)) == NULL
      || strlen(buf) != 2 * sizeof(LINK)) {
    return NULL;
  }

  result->tag = FILENAME_LINK;
  target = &(result->link);

  for (i = 0; i < sizeof(LINK); i++) {
    char s[] = {buf[i*2], buf[i*2+1], '\0'};
    sscanf(s, "%x", &v);
    target[i] = v;
  }

  return 1;
}

/*
 * Lines in the host keys file are of the form
 * 
 *   type@port:hostname keydata
 * 
 * e.g.
 * 
 *   rsa@22:foovax.example.org 0x23,0x293487364395345345....2343
 */

int verify_host_key(const char *hostname, int port,
		    const char *keytype, const char *key)
{
    FILE *fp;
    TC fname_self[] = {TC_FOWN, TNULL};
    TC fname_fn[] = {TCS_FNAME_HOSTKEYS};
    struct Filename fn;
    char *line;
    int ret;

    /*
     * Indirect file reference is used here.
     * Be sure to chg_wrk beforehand.
     */
    fn.tag = FILENAME_LINKNAME;
    get_lnk(fname_self, &(fn.link), F_NORM);
    tc_strcpy(fn.path, fname_fn);
    fp = f_open(fn, "r", 0);
    if (!fp)
	return 1;		       /* key does not exist */

    ret = 1;
    while ( (line = fgetline(fp)) ) {
	int i;
	char *p = line;
	char porttext[20];

	line[strcspn(line, "\n")] = '\0';   /* strip trailing newline */

	i = strlen(keytype);
	if (strncmp(p, keytype, i))
	    goto done;
	p += i;

	if (*p != '@')
	    goto done;
	p++;

	sprintf(porttext, "%d", port);
	i = strlen(porttext);
	if (strncmp(p, porttext, i))
	    goto done;
	p += i;

	if (*p != ':')
	    goto done;
	p++;

	i = strlen(hostname);
	if (strncmp(p, hostname, i))
	    goto done;
	p += i;

	if (*p != ' ')
	    goto done;
	p++;

	/*
	 * Found the key. Now just work out whether it's the right
	 * one or not.
	 */
	if (!strcmp(p, key))
	    ret = 0;		       /* key matched OK */
	else
	    ret = 2;		       /* key mismatch */

	done:
	sfree(line);
	if (ret != 1)
	    break;
    }

    fclose(fp);
    return ret;
}

void store_host_key(const char *hostname, int port,
		    const char *keytype, const char *key)
{
    TC fname_self[] = {TC_FOWN, TNULL};
    TC fname_fntmp[] = {TCS_FNAME_HOSTKEYSTMP};
    TC fname_fn[] = {TCS_FNAME_HOSTKEYS};
    struct Filename fn_tmp, fn;
    FILE *rfp, *wfp;
    char *newtext, *line;
    int headerlen;
    char filename[FNLEN], tmpfilename[FNLEN];
    LINK ln_self, ln1, ln2;
    W err;

    newtext = dupprintf("%s@%d:%s %s\n", keytype, port, hostname, key);
    headerlen = 1 + strcspn(newtext, " ");   /* count the space too */

    /*
     * Open both the old file and a new file.
     *
     * Indirect file reference is used here.
     * Be sure to chg_wrk beforehand.
     */
    fn_tmp.tag = FILENAME_LINKNAME;
    get_lnk(fname_self, &(fn_tmp.link), F_NORM);
    tc_strcpy(fn_tmp.path, fname_fntmp);

    wfp = f_open(fn_tmp, "w", 0);
    if (!wfp)
      goto err_ret;

    fn.tag = FILENAME_LINKNAME;
    get_lnk(fname_self, &(fn.link), F_NORM);
    tc_strcpy(fn.path, fname_fn);
    rfp = f_open(fn, "r", 0);

    /*
     * Copy all lines from the old file to the new one that _don't_
     * involve the same host key identifier as the one we're adding.
     */
    if (rfp) {
        while ( (line = fgetline(rfp)) ) {
            if (strncmp(line, newtext, headerlen))
                fputs(line, wfp);
        }
        fclose(rfp);
    }

    /*
     * Now add the new line at the end.
     */
    fputs(newtext, wfp);

    fclose(wfp);

    /* Here we do the renaming in BTRON: 
     * Delete the old one first, and then change the name of the new one */
    get_lnk(fname_self, &ln_self, F_NORM);
    err = get_lnk(fname_fn, &ln1, F_NORM);
    if (err >= 0) {
      err = del_fil(&ln_self, &ln1, 1);
      if (err < 0) {
        goto err_ret;
      }
    }
    err = get_lnk(fname_fntmp, &ln2, F_NORM);
    if (err < 0) {
      goto err_ret;
    }
    chg_fnm(&ln2, fname_fn);

err_ret:
    sfree(newtext);
}

void read_random_seed(noise_consumer_t consumer)
{
  TC fname_randseed[] = {TCS_FNAME_RANDSEED};
  LINK lnk;
  W err, fd;
  B buf[1024];
  W offset;
  W rec_size, read_size;

  /*
   * Indirect file reference is used here.
   * Be sure to chg_wrk beforehand.
   */
  err = get_lnk(fname_randseed, &lnk, F_NORM);
  if (err < 0) {
    return;
  }

  fd = opn_fil(&lnk, F_READ, NULL);
  if (fd < 0) {
    return;
  }

  err = fnd_rec(fd, F_TOPEND, 1<<31, 0, NULL);
  if (err < 0) {
    goto label_ret;
  }

  err = rea_rec(fd, 0, NULL, 0, &rec_size, NULL);
  if (err < 0) {
    goto label_ret;
  }

  for (offset = 0; offset < rec_size; ) {
    read_size = rec_size - offset;
    if (read_size > sizeof(buf)) {
      read_size = sizeof(buf);
    }
    err = rea_rec(fd, offset, buf, read_size, NULL, NULL);
    if (err < 0) {
      goto label_ret;
    }    
    consumer(buf, read_size);
    offset += read_size;
  }

label_ret:
  if (fd > 0) {
    cls_fil(fd);
  }
}

void write_random_seed(void *data, int len)
{
  TC fname_randseed[] = {TCS_FNAME_RANDSEED};
  TC fname_self[] = {TC_FOWN, TNULL};
  LINK lnk;
  W err, fd = -1;

  /*
   * Indirect file reference is used here.
   * Be sure to chg_wrk beforehand.
   */
  err = get_lnk(fname_randseed, &lnk, F_NORM);
  if (err >= 0) {
    fd = opn_fil(&lnk, F_WRITE, NULL);
    if (fd < 0) {
      return;
    }
  }
  else {
    err = get_lnk(fname_self, &lnk, F_NORM);
    if (err < 0) {
      return;
    }
    fd = cre_fil(&lnk, fname_randseed, NULL, 31, F_FIX);
    if (fd < 0) {
      return;
    }
  }

  err = fnd_rec(fd, F_TOPEND, 1<<31, 0, NULL);
  if (err < 0) {
    apd_rec(fd, (B*)data, len, 31, 0, 0);
    goto label_ret;
  }

  /*
   * Don't truncate the random seed file if it already exists; if
   * something goes wrong half way through writing it, it would
   * be better to leave the old data there than to leave it empty.
   */
  wri_rec(fd, 0, (B*)data, len, NULL, NULL, 0);

label_ret:
  if (fd > 0) {
    cls_fil(fd);
  }
}

