/*
 * btwin.c: the main code that runs a PuTTY terminal emulator and
 * backend in a BTRON window.
 */

#include <basic.h>
#include <bstdio.h>
#include <bstdlib.h>
#include <bassert.h>
#include <stdarg.h>
#include <btron/proctask.h>
#include <btron/event.h>
#include <btron/hmi.h>
#include <btron/dp.h>
#include <btron/vobj.h>
#include <keycode.h>
#include <tcode.h>
#include <tstring.h>
#include <btron/libapp.h>
#include <btron/tf.h>
#include <tad.h>
#include <btron/taddata.h>

#define PUTTY_DO_GLOBALS	       /* actually _define_ globals */

#include "putty.h"
#include "terminal.h"
#include "tcsliteral.h"

extern DEV_SPEC SCREEN;

#define CAT2(x,y) x ## y
#define CAT(x,y) CAT2(x,y)
#define ASSERT(x) enum {CAT(assertion_,__LINE__) = 1 / (x)}

/* Colours come in two flavours: configurable, and xterm-extended. */
#define NCFGCOLOURS (lenof(((Config *)0)->colours))
#define NEXTCOLOURS 240 /* 216 colour-cube plus 24 shades of grey */
#define NALLCOLOURS (NCFGCOLOURS + NEXTCOLOURS)

void repaint(void* frontend);
void on_redisp(WEVENT *ev, void *frontend);

#define MAIN_WINDOW_ATTR (WA_WORK | WA_SIZE | WA_VHDL | WA_HHDL | WA_RBAR | WA_BGDSP | WA_THIN)

typedef struct {
  WID wid;
} Window;

struct gui_data {
  Window* window;
  BMP* pixmap;
  struct draw_ctx *ctx;
  CSPEC cspec;
  COLOR cols[NALLCOLOURS];
  char wintitle[sizeof(((Config *)0)->wintitle)];
  int busy_status;
  int font_width;
  int font_height;
  int width;
  int height;
  int direct_to_font;
  Backend *back;
  void *backhandle;
  Terminal *term;
  Config cfg;
  struct unicode_data ucsdata;
  void *ldisc;
  int exited;
  MESSAGE *msg;
  putty_fusen *fusen;
  WDSTAT rsave;
  int wfull;
  /* mouse_x, mouse_y are to be used for motion_event */
  int mouse_x;
  int mouse_y;
  int window_moving;
};

struct draw_ctx {
  W wgid;
  W bgid;
  struct gui_data *inst;
};

static int send_raw_mouse;

extern tree234 *fds;

void ldisc_update(void *frontend, int echo, int edit)
{
    /*
     * This is a stub in pterm. If I ever produce a Unix
     * command-line ssh/telnet/rlogin client (i.e. a port of plink)
     * then it will require some termios manoeuvring analogous to
     * that in the Windows plink.c, but here it's meaningless.
     */
}

char *get_ttymode(void *frontend, const char *mode)
{
    struct gui_data *inst = (struct gui_data *)frontend;
    return term_get_ttymode(inst->term, mode);
}

int from_backend(void *frontend, int is_stderr, const char *data, int len)
{
    struct gui_data *inst = (struct gui_data *)frontend;
    return term_data(inst->term, is_stderr, data, len);
}

int from_backend_untrusted(void *frontend, const char *data, int len)
{
    struct gui_data *inst = (struct gui_data *)frontend;
    return term_data_untrusted(inst->term, data, len);
}

int get_userpass_input(prompts_t *p, unsigned char *in, int inlen)
{
    struct gui_data *inst = (struct gui_data *)p->frontend;
    int ret;
#if 0 /* BTRON */
    ret = cmdline_get_passwd_input(p, in, inlen);
    if (ret == -1)
#endif
	ret = term_get_userpass_input(inst->term, p, in, inlen);
    return ret;
}

void logevent(void *frontend, const char *s)
{
  puts((char *)s);
}

void log_packet(void *handle, int direction, int type,
		char *texttype, void *data, int len,
		int n_blanks, const struct logblank_t *blanks)
{
  ; /* no-op */
}

/*
 * Timers are global across all sessions (even if we were handling
 * multiple sessions, which we aren't), so the current timer ID is
 * a global variable.
 */
typedef struct {
  long ticks;
  long next;
} timer_task_param;

static volatile W timer_tskid = -1;
static volatile timer_task_param timer_param;

static void timer_trigger_task(W arg)
{
  long now;

  for (;;) {
    do {
      now = timer_param.next;
      slp_tsk(timer_param.ticks);
    } while (now != timer_param.next);

    if (run_timers(now, &(timer_param.next))) {
      timer_param.ticks = timer_param.next - GETTICKCOUNT();
      if (timer_param.ticks <= 0) {
        timer_param.ticks = 1;
      }
    }
    else {
      slp_tsk(-1);
    }
  }

  /* not reached */
  ext_tsk();
}

void timer_change_notify(long next)
{
  timer_param.next = next;
  timer_param.ticks = next - GETTICKCOUNT();
  if (timer_param.ticks <= 0) {
    timer_param.ticks = 1;
  }

  if (timer_tskid < 0) {
    P_STATE pst;
    prc_sts(0, &pst, NULL);
    timer_tskid = cre_tsk((FP)timer_trigger_task, pst.priority, 0);
  }
  else {
    wup_tsk(timer_tskid);
  }
}

void notify_remote_exit(void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  int exitcode;

  if (!inst->exited &&
      (exitcode = inst->back->exitcode(inst->backhandle)) >= 0) {
    inst->exited = TRUE;
    if (inst->cfg.close_on_exit == FORCE_ON ||
        (inst->cfg.close_on_exit == AUTO && exitcode == 0))
#if 0
      gtk_main_quit();	       /* just go */
#else
      ;
#endif
    if (inst->ldisc) {
      ldisc_free(inst->ldisc);
      inst->ldisc = NULL;
    }
    if (inst->back) {
      inst->back->free(inst->backhandle);
      inst->backhandle = NULL;
      inst->back = NULL;
      term_provide_resize_fn(inst->term, NULL, NULL);
      update_specials_menu(inst);
    }
#if 0
    gtk_widget_show(inst->restartitem);
#endif
  }
}

void frontend_keypress(void *handle)
{
  struct gui_data *inst = (struct gui_data *)handle;

  /*
   * If our child process has exited but not closed, terminate on
   * any keypress.
   */
  /* TODO: add appropreate finalizations, e.g. oend_prc */
  if (inst->exited) {
    oend_prc(((M_EXECREQ*)(inst->msg))->vid, inst->fusen, 0);
    ext_prc(0);
  }
}

void update_specials_menu(void *frontend)
{
  ; /* TODO */
}

/*
 * Return the top-level GtkWindow associated with a particular
 * front end instance.
 */
void *get_window(void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  return inst->window;
}

/*
 * Minimise or restore the window in response to a server-side
 * request.
 */
void set_iconic(void *frontend, int iconic)
{
  ; /* BTRON cannot do this */
}

/*
 * Move the window in response to a server-side request.
 */
void move_window(void *frontend, int x, int y)
{
  W err;
  RECT r;
  struct gui_data *inst = (struct gui_data *)frontend;

  r.c.left = r.c.right  = x;
  r.c.top  = r.c.bottom = y;
  wcnv_rec((MAIN_WINDOW_ATTR|WA_FRAME) & ~WA_WORK, &r, MAIN_WINDOW_ATTR & (WA_WORK|WA_FRAME));

  err = wmov_wnd(inst->window->wid, &r);
  if (err >= 0) {
    inst->wfull = 0;
    if (err >= 1) {
      on_redisp(NULL, frontend);
    }
  }
}

/*
 * Move the window to the top or bottom of the z-order in response
 * to a server-side request.
 */
void set_zorder(void *frontend, int top)
{
  ; /* BTRON cannot do this */
}

/*
 * Refresh the window in response to a server-side request.
 */
void refresh_window(void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  term_invalidate(inst->term);
}

/*
 * Maximise or restore the window in response to a server-side
 * request.
 */
void set_zoomed(void *frontend, int zoomed)
{
  W err = 0;
  struct gui_data *inst = (struct gui_data *)frontend;

  wget_sts(inst->window->wid, &(inst->rsave), NULL);
  if (inst->wfull) {
    if (!zoomed) {
      err = wchg_wnd(inst->window->wid, NULL, W_MOVE);
      if (err >= 0) {
        inst->wfull = (err == 1);
      }
    }
  }
  else {
    if (zoomed) {
      err = wchg_wnd(inst->window->wid, NULL, W_MOVE);
      if (err >= 0) {
        inst->wfull = (err == 1);
      }
    }
  }

  if (err >= 1) {
    on_redisp(NULL, frontend);
  }
}

/*
 * Report whether the window is iconic, for terminal reports.
 */
int is_iconic(void *frontend)
{
  return FALSE;
}

/*
 * Report the window's position, for terminal reports.
 */
void get_window_pos(void *frontend, int *x, int *y)
{
  WDSTAT stat;
  struct gui_data *inst = (struct gui_data *)frontend;

  stat.attr = WA_FRAME;
  wget_sts(inst->window->wid, &stat, NULL);

  *x = stat.r.c.left;
  *y = stat.r.c.top;
}

/*
 * Report the window's pixel size, for terminal reports.
 */
void get_window_pixels(void *frontend, int *x, int *y)
{
  WDSTAT stat;
  struct gui_data *inst = (struct gui_data *)frontend;

  stat.attr = WA_WORK;
  wget_sts(inst->window->wid, &stat, NULL);

  *x = stat.r.c.right - stat.r.c.left + 1;
  *y = stat.r.c.bottom - stat.r.c.right + 1;
}

/*
 * Return the window or icon title.
 */
char *get_window_title(void *frontend, int icon)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  return inst->wintitle;
}

static void update_mouseptr(struct gui_data *inst)
{
  switch (inst->busy_status) {
      case BUSY_NOT:
        /* We never make pointers invisible in BTRON */
        gset_ptr(PS_SELECT, NULL, (COLOR)-1, (COLOR)-1);
      case BUSY_WAITING:    /* XXX can we do better? */
      case BUSY_CPU:
	/* We always display these cursors. */
        gset_ptr(PS_BUSY, NULL, (COLOR)-1, (COLOR)-1);
	break;
      default:
	assert(0);
  }
}

/*
 * set or clear the "raw mouse message" mode
 */
void set_raw_mouse_mode(void *frontend, int activate)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  activate = activate && !inst->cfg.no_mouse_rep;
  send_raw_mouse = activate;
  update_mouseptr(inst);
}

void set_busy_status(void *frontend, int status)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  inst->busy_status = status;
  update_mouseptr(inst);
}

void set_title(void *frontend, char *title)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  TC tc_wintitle[sizeof(((Config *)0)->wintitle) + 1];
  if (inst->window == NULL) {
    return;
  }
  strcpy(inst->wintitle, title);
  eucstotcs(tc_wintitle, title);
  wset_tit(inst->window->wid, -1, tc_wintitle, 0);
}

void set_icon(void *frontend, char *title)
{
  ; /* nop */
}

void set_sbar(void *frontend, int total, int start, int page)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  W rbar;
  W val[4];
  if (!inst->cfg.scrollbar || inst->window == NULL)
    return;
  wget_bar(inst->window->wid, &rbar, NULL, NULL);
  val[0] = start; /* clo */
  val[1] = start+page; /* chi */
  val[2] = 0; /* lo */
  val[3] = total; /* hi */
  cset_val(rbar, 4, val);
}

void sys_cursor(void *frontend, int x, int y)
{
  /*
   * This is meaningless under X.
   */
}

void do_beep(void *frontend, int mode)
{
  if (mode != BELL_VISUAL) {
    sig_buz(0);
  }
}

int char_width(Context ctx, int uc)
{
    /*
     * Under X, any fixed-width font really _is_ fixed-width.
     * Double-width characters will be dealt with using a separate
     * font. For the moment we can simply return 1.
     */
    return 1;
}

Context get_ctx(void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;

  if (!inst->window)
    return NULL;

  return inst->ctx;
}

void free_ctx(Context ctx)
{
  ; /* nop */
}

void connection_fatal(void *frontend, char *p, ...)
{
  struct gui_data *inst = (struct gui_data *)frontend;

  va_list ap;
  char *msg;
  va_start(ap, p);
  msg = dupvprintf(p, ap);
  va_end(ap);
  inst->exited = TRUE;

  fatal_message_box(inst->window, msg);

  sfree(msg);
  if (inst->cfg.close_on_exit == FORCE_ON)
    cleanup_exit(1);
}

static volatile TF_CTX tfctx = NULL;
static volatile W tfsem = -1;

static TF_CTX get_tf_ctx()
{
  if (tfsem < 0) {
    tfsem = cre_sem(1, SEM_EXCL|DELEXIT);
  }
  if (tfctx == NULL) {
    wai_sem(tfsem, T_FOREVER);
    tf_open_ctx(&tfctx);
    tf_set_options(tfctx, tf_to_id(TF_ID_OPT_CONVERT, "CONV_SBCHAR_HAN"), 0);
    tf_set_options(tfctx, tf_to_id(TF_ID_OPT_CONVERT, "UNDEFINED_CHAR"), 0);
    /* BTRON TODO: set this according to inst->cfg */
    tf_set_profile(tfctx, tf_to_id(TF_ID_PROFSET_CONVERTFROM, "EUC-JP"));
    sig_sem(tfsem);
  }
  return tfctx;
}

WERR gdra_stp_wc(GID gid, W x, W y, char *str, W len, DCM mode, FSSPEC* font)
{
  int i;
  int font_width_save = font->size.h;
  TC buf[1024];
  W buflen;
  TF_CTX ctx = get_tf_ctx();
  int zenkaku = TRUE;
  W err;
  int first = TRUE;

  font->size.h = font_width_save * 2;
  gset_fon(gid, font);

  do {
    buflen = sizeof(buf)/sizeof(buf[0]);
    wai_sem(tfsem, T_FOREVER);
    err = tf_strtotcs(ctx, (first ? str : NULL), len,
                      (first ? TF_ATTR_START : 0)|TF_ATTR_CONT, buf, &buflen);
    sig_sem(tfsem);
    first = FALSE;

    for (i = 0; i < buflen; ) {
      if (buf[i] == TNULL) {
        break;
      }
      else if ((buf[i] & 0xff00) == 0xff00) {
        TT_RATIO *ttr = &(buf[i+2]);
        if (ttr->subid == TT_RATIO_SID) {
          zenkaku = (ttr->w_ratio == 0);
        }
        i += 2 + (buf[i+1] & 0xffff) / 2;
        font->size.h = font_width_save;
        if (zenkaku) {
          font->size.h *= 2;
        }
        gset_fon(gid, font);
      }
      else {
        gdra_chp(gid, x, y, buf[i], mode);
        x += font->size.h;
        i++;
      }
    }
  }
  while (err > 0);

  gdra_chr(gid, (0xfe00|TSC_SYS), mode);

  font->size.h = font_width_save;
  return 0;
}

/*
 * Draw a line of text in the window, at given character
 * coordinates, in given attributes.
 *
 * We are allowed to fiddle with the contents of `text'.
 */
void do_text_internal(Context ctx, int x, int y, wchar_t *text, int len,
		      unsigned long attr, int lattr)
{
  struct draw_ctx *dctx = (struct draw_ctx *)ctx;
  struct gui_data *inst = dctx->inst;
  GID bgid = dctx->bgid;
  int ncombining, combining;
  int nfg, nbg, t, fontid, shadow, rlen, widefactor;
  int monochrome = ((SCREEN.attr & 0x7) == 0);
  COLOR fgcol, bgcol;
  RECT rrec;
  PAT pat;
  FSSPEC font;
  int ascent;

  if (attr & TATTR_COMBINING) {
    ncombining = len;
    len = 1;
  } else
    ncombining = 1;

  nfg = ((monochrome ? ATTR_DEFFG : (attr & ATTR_FGMASK)) >> ATTR_FGSHIFT);
  nbg = ((monochrome ? ATTR_DEFBG : (attr & ATTR_BGMASK)) >> ATTR_BGSHIFT);
  if (!!(attr & ATTR_REVERSE) ^ (monochrome && (attr & TATTR_ACTCURS))) {
    t = nfg;
    nfg = nbg;
    nbg = t;
  }
  if (inst->cfg.bold_colour && (attr & ATTR_BOLD)) {
    if (nfg < 16) nfg |= 8;
    else if (nfg >= 256) nfg |= 1;
  }
  if (inst->cfg.bold_colour && (attr & ATTR_BLINK)) {
    if (nbg < 16) nbg |= 8;
    else if (nbg >= 256) nbg |= 1;
  }
  if ((attr & TATTR_ACTCURS) && !monochrome) {
    nfg = 260;
    nbg = 261;
  }

  fontid = shadow = 0;

  if (attr & ATTR_WIDE) {
    widefactor = 2;
    fontid |= 2;
  } else {
    widefactor = 1;
  }

  if ((attr & ATTR_BOLD) && !inst->cfg.bold_colour) {
//    if (inst->fonts[fontid | 1])
    fontid |= 1;
//    else
//      shadow = 1;
  }

  if ((lattr & LATTR_MODE) != LATTR_NORM) {
    x *= 2;
    if (x >= inst->term->cols)
      return;
    if (x + len*2*widefactor > inst->term->cols)
      len = (inst->term->cols-x)/2/widefactor;/* trim to LH half */
    rlen = len * 2;
  } else
    rlen = len;

  {
    GRGN rclp;
    rclp.r.c.left = x*inst->font_width+inst->cfg.window_border;
    rclp.r.c.top = y*inst->font_height+inst->cfg.window_border;
    rclp.r.c.right = rclp.r.c.left + rlen*widefactor*inst->font_width;
    rclp.r.c.bottom = rclp.r.c.top + inst->font_height;
    rclp.ns = 0;
    gset_clp(bgid, &rclp);
  }

//  gdk_gc_set_foreground(gc, &inst->cols[nbg]);
  fgcol = inst->cols[nbg];
  rrec.c.left = x*inst->font_width+inst->cfg.window_border;
  rrec.c.top = y*inst->font_height+inst->cfg.window_border;
  rrec.c.right = rrec.c.left + rlen*widefactor*inst->font_width + 10;
  rrec.c.bottom = rrec.c.top + inst->font_height + 10;
  pat.spat.kind = 0;
  pat.spat.hsize = pat.spat.vsize = 1;
  pat.spat.fgcol = pat.spat.bgcol = fgcol;
  pat.spat.mask = FILL100;
  gfil_rec(bgid, rrec, &pat, 0, G_STORE);

//  gdk_gc_set_foreground(gc, &inst->cols[nfg]);
  fgcol = inst->cols[nfg];
  bgcol = inst->cols[258];
  gset_chc(bgid, fgcol, bgcol | 0x80000000); /* transparent */

  /* TODO: use configuration info */
  font.name[0] = TNULL;
  font.fclass = FTC_DEFAULT;
  font.attr = FT_REGULAR | FT_MIDI | FT_RIGHT | FT_SOLID 
    | FT_HDRAW | FT_FIXED;
  font.size.h = inst->font_width;
  font.size.v = inst->font_height;
  ascent = font.size.v - 1;

  {
    /* TODO: optimize here to directly convert wcs into tcs */
    char *gcs;
    wchar_t *wcs;
    int i;
    int gcslen;

    wcs = snewn(len*ncombining+1, wchar_t);
    for (i = 0; i < len*ncombining; i++) {
      wcs[i] = text[i];
    }

    gcs = snewn(len+1, char);
    for (combining = 0; combining < ncombining; combining++) {
      /* BTRON TODO */
      wc_to_mb(CS_ISO8859_1, 0,
               wcs + combining, len, gcs, len, ".", NULL, NULL);
      gcslen = strlen(gcs);

      gdra_stp_wc(bgid, x*inst->font_width+inst->cfg.window_border, 
                  y*inst->font_height+inst->cfg.window_border+ascent,
                  gcs, -1, G_STORE, &font);
      if (shadow)
        gdra_stp_wc(bgid, x*inst->font_width+inst->cfg.window_border+inst->cfg.shadowboldoffset, 
                    y*inst->font_height+inst->cfg.window_border+ascent,
                    gcs, -1, G_STORE, &font);
    }
    sfree(gcs);
    sfree(wcs);
  }

  if (attr & ATTR_UNDER) {
#if 0 /* TODO */
    int uheight = inst->fonts[0]->ascent + 1;
    if (uheight >= inst->font_height)
      uheight = inst->font_height - 1;
#else
    int uheight = ascent;
#endif
    PNT p1, p2;
    p1.x = x*inst->font_width+inst->cfg.window_border;
    p1.y = y*inst->font_height + uheight + inst->cfg.window_border;
    p2.x = (x+len)*widefactor*inst->font_width-1+inst->cfg.window_border;
    p2.y = y*inst->font_height + uheight + inst->cfg.window_border;
    pat.spat.fgcol = pat.spat.bgcol = fgcol;
    gdra_lin(bgid, p1, p2, ((LN_SOLID << 8) | 1), &pat, G_STORE);
  }

  if ((lattr & LATTR_MODE) != LATTR_NORM) {
    /*
     * I can't find any plausible StretchBlt equivalent in the
     * X server, so I'm going to do this the slow and painful
     * way. This will involve repeated calls to
     * gdk_draw_pixmap() to stretch the text horizontally. It's
     * O(N^2) in time and O(N) in network bandwidth, but you
     * try thinking of a better way. :-(
     */
    /* Fortunately, we have StretchBlt equivalent in BTRON :-) */
    RECT sr, dr;
    dr.c.left = sr.c.left = x*inst->font_width+inst->cfg.window_border;
    dr.c.top = sr.c.top = y*inst->font_height+inst->cfg.window_border;
    sr.c.right = sr.c.left + len * widefactor * inst->font_width;
    dr.c.right = dr.c.left + 2 * len * widefactor * inst->font_width;
    dr.c.bottom = sr.c.bottom = sr.c.top + inst->font_height;
    if ((lattr & LATTR_MODE) != LATTR_WIDE) {
      /* Now stretch vertically, in the same way. */
      if ((lattr & LATTR_MODE) == LATTR_BOT) {
        sr.c.top = (sr.c.top + sr.c.bottom) / 2;
      }
      else {
        sr.c.bottom = (sr.c.top + sr.c.bottom) / 2;
      }
    }
    gcop_bmp(bgid, &sr, bgid, &dr, NULL, G_STORE);
  }
}

void do_text(Context ctx, int x, int y, wchar_t *text, int len,
	     unsigned long attr, int lattr)
{
  struct draw_ctx *dctx = (struct draw_ctx *)ctx;
  struct gui_data *inst = dctx->inst;
  GID wgid = dctx->wgid;
  int widefactor;
  RECT r;

  do_text_internal(ctx, x, y, text, len, attr, lattr);

  if (attr & ATTR_WIDE) {
    widefactor = 2;
  } else {
    widefactor = 1;
  }

  if ((lattr & LATTR_MODE) != LATTR_NORM) {
    x *= 2;
    if (x >= inst->term->cols)
      return;
    if (x + len*2*widefactor > inst->term->cols)
      len = (inst->term->cols-x)/2/widefactor;/* trim to LH half */
    len *= 2;
  }

  r.c.left = x*inst->font_width+inst->cfg.window_border;
  r.c.top = y*inst->font_height+inst->cfg.window_border;
  r.c.right = r.c.left + len*widefactor*inst->font_width;
  r.c.bottom = r.c.top + inst->font_height;
  
  /* When the window is moving, gid is locked and cannot be painted */
  if (!inst->window_moving) {
    grst_bmp(wgid, NULL, inst->pixmap, &r, NULL, G_STORE);
  }
}

void do_cursor(Context ctx, int x, int y, wchar_t *text, int len,
	       unsigned long attr, int lattr)
{
  struct draw_ctx *dctx = (struct draw_ctx *)ctx;
  struct gui_data *inst = dctx->inst;
  GID bgid = dctx->bgid;
  GID wgid = dctx->wgid;
  COLOR fgcol;
  PAT pat;
  RECT r;
  PNT p1, p2;

  int active, passive, widefactor;

  if (attr & TATTR_PASCURS) {
    attr &= ~TATTR_PASCURS;
    passive = 1;
  } else
    passive = 0;
  if ((attr & TATTR_ACTCURS) && inst->cfg.cursor_type != 0) {
    attr &= ~TATTR_ACTCURS;
    active = 1;
  } else
    active = 0;
  do_text_internal(ctx, x, y, text, len, attr, lattr);

  if (attr & TATTR_COMBINING)
    len = 1;

  if (attr & ATTR_WIDE) {
    widefactor = 2;
  } else {
    widefactor = 1;
  }

  if ((lattr & LATTR_MODE) != LATTR_NORM) {
    x *= 2;
    if (x >= inst->term->cols)
      return;
    if (x + len*2*widefactor > inst->term->cols)
      len = (inst->term->cols-x)/2/widefactor;/* trim to LH half */
    len *= 2;
  }

  pat.spat.kind = 0;
  pat.spat.hsize = pat.spat.vsize = 1;
  pat.spat.mask = FILL100;

  if (inst->cfg.cursor_type == 0) {
    /*
     * An active block cursor will already have been done by
     * the above do_text call, so we only need to do anything
     * if it's passive.
     */
    if (passive) {
      r.c.left = x*inst->font_width+inst->cfg.window_border;
      r.c.top = y*inst->font_height+inst->cfg.window_border;
      r.c.right = r.c.left + len*widefactor*inst->font_width - 1;
      r.c.bottom = r.c.top + inst->font_height - 1;
      fgcol = inst->cols[261];
      pat.spat.fgcol = pat.spat.bgcol = fgcol;
      gfil_rec(bgid, r, &pat, 0, G_STORE);
    }
  } else {
    int uheight;
    int startx, starty, dx, dy, length, i;
    int char_width;
    int ascent = inst->font_height - 1; /* TODO? */

    if ((attr & ATTR_WIDE) || (lattr & LATTR_MODE) != LATTR_NORM)
      char_width = 2*inst->font_width;
    else
      char_width = inst->font_width;

    if (inst->cfg.cursor_type == 1) {
//      uheight = inst->fonts[0]->ascent + 1;
      uheight = ascent;
      if (uheight >= inst->font_height)
        uheight = inst->font_height - 1;

      startx = x * inst->font_width + inst->cfg.window_border;
      starty = y * inst->font_height + inst->cfg.window_border + uheight;
      dx = 1;
      dy = 0;
      length = len * widefactor * char_width;
    } else {
      int xadjust = 0;
      if (attr & TATTR_RIGHTCURS)
        xadjust = char_width - 1;
      startx = x * inst->font_width + inst->cfg.window_border + xadjust;
      starty = y * inst->font_height + inst->cfg.window_border;
      dx = 0;
      dy = 1;
      length = inst->font_height;
    }

    fgcol = inst->cols[261];
    pat.spat.fgcol = pat.spat.bgcol = fgcol;
    if (passive) {
      for (i = 0; i < length; i++) {
        if (i % 2 == 0) {
          p1.x = startx; p1.y = starty;
          gdra_pnt(bgid, p1, fgcol, G_STORE);
        }
        startx += dx;
        starty += dy;
      }
    } else if (active) {
      p1.x = startx; p1.y = starty;
      p2.x = startx + (length-1) * dx;
      p2.y = starty + (length-1) * dy;
      gdra_lin(bgid, p1, p2, lattr, &pat, G_STORE);
    } /* else no cursor (e.g., blinked off) */
  }

  r.c.left = x*inst->font_width+inst->cfg.window_border;
  r.c.top = y*inst->font_height+inst->cfg.window_border;
  r.c.right = r.c.left + len*widefactor*inst->font_width;
  r.c.bottom = r.c.top + inst->font_height;

  /* When the window is moving, gid is locked and cannot be painted */
  if (!inst->window_moving) {
    grst_bmp(wgid, NULL, inst->pixmap, &r, NULL, G_STORE);
  }
}

void cmdline_error(char *p, ...)
{
  va_list ap;
  printf("%s: ", appname);
  va_start(ap, p);
  vprintf(p, ap);
  va_end(ap);
  putchar('\n');

  /* TODO: do appropreate finalizations, e.g. oend_prc */
  ext_prc(1);
}

void draw_backing_rect(struct gui_data *inst)
{
  W gid;
  PAT pat;

  if (inst->ctx == NULL) {
    return;
  }
  gid = inst->ctx->bgid;
  gset_clp(gid, NULL);

  pat.spat.kind = 0;
  pat.spat.hsize = pat.spat.vsize = 1;
  pat.spat.fgcol = pat.spat.bgcol = inst->cols[258];
  pat.spat.mask = FILL100;
  gfil_rec(gid, inst->pixmap->bounds, &pat, 0, G_STORE);
}

void request_resize(void *frontend, int w, int h)
{
  W err;
  WDSTAT stat;
  struct gui_data *inst = (struct gui_data *)frontend;
  int need_size = FALSE;

  /* First, resize & update the screen bitmap */

  if (inst->pixmap == NULL) {
    /* singleton; no need to be freed */
    struct draw_ctx *dctx;
    inst->pixmap = (BMP*)smalloc(sizeof(BMP) + sizeof(UB*) * SCREEN.planes);
    inst->pixmap->planes = SCREEN.planes;
    inst->pixmap->pixbits = SCREEN.pixbits;
    inst->pixmap->bounds.c.left = 0;
    inst->pixmap->bounds.c.top = 0;
    inst->pixmap->bounds.c.right = (w+1) * inst->font_width + inst->cfg.window_border * 2 + 1;
    inst->pixmap->bounds.c.bottom = (h+1) * inst->font_height + inst->cfg.window_border * 2 + 1;
    bmalloc(inst->pixmap); /* rowbytes, baseaddr are set here */
    dctx = snew(struct draw_ctx);
    dctx->inst = inst;
    dctx->wgid = wget_gid(inst->window->wid);
    dctx->bgid = gopn_mem(NULL, inst->pixmap, (B*)&(inst->cspec));
    inst->ctx = dctx;
    need_size = TRUE;
  }
  else if (w != inst->width || h != inst->height) {
    gcls_env(inst->ctx->bgid);
    bfree(inst->pixmap);
    inst->pixmap->bounds.c.right = (w+1) * inst->font_width + inst->cfg.window_border * 2 + 1;
    inst->pixmap->bounds.c.bottom = (h+1) * inst->font_height + inst->cfg.window_border * 2 + 1;
    err = bmalloc(inst->pixmap); /* rowbytes, baseaddr are set here */
    inst->ctx->bgid = gopn_mem(NULL, inst->pixmap, (B*)&(inst->cspec));
    need_size = TRUE;
  }

  draw_backing_rect(inst);

  if (need_size) {
    inst->cfg.width = inst->width = w;
    inst->cfg.height = inst->height = h;
    if (inst->term) {
      term_size(inst->term, h, w, inst->cfg.savelines);
    }
  }

  if (inst->term) {
    term_invalidate(inst->term);
  }

  /* Second, change the window size */

  stat.attr = WA_WORK;
  wget_sts(inst->window->wid, &stat, NULL);

  if (!inst->wfull) {
    stat.r.c.right = stat.r.c.left + w * inst->font_width + inst->cfg.window_border * 2;
    stat.r.c.bottom = stat.r.c.top + h * inst->font_height + inst->cfg.window_border * 2;
    wcnv_rec((MAIN_WINDOW_ATTR|WA_WORK) & ~WA_FRAME, &(stat.r), 
             MAIN_WINDOW_ATTR & (WA_WORK|WA_FRAME));
    err = wrsz_wnd(inst->window->wid, &(stat.r));
    if (err >= 0) {
      inst->wfull = 0;
      if (err >= 1) {
        on_redisp(NULL, frontend);
      }
    }
  }
  repaint(frontend);
}

void palette_set(void *frontend, int n, int r, int g, int b)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  if (n >= 16)
    n += 256 - 16;
  if (n > NALLCOLOURS)
    return;
  inst->cols[n] = 0x10000000
    | ((r & 0xff) << 16)
    | ((g & 0xff) << 8)
    | (b & 0xff);

  if (n == 258) {
    /* Default Background changed. Ensure space between text area and
     * window border is redrawn */
    draw_backing_rect(inst);
    repaint(frontend);
  }
}

void palette_reset(void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  /* This maps colour indices in inst->cfg to those used in inst->cols. */
  static const int ww[] = {
    256, 257, 258, 259, 260, 261,
    0, 8, 1, 9, 2, 10, 3, 11,
    4, 12, 5, 13, 6, 14, 7, 15
  };
  int i;

  assert(lenof(ww) == NCFGCOLOURS);

  for (i = 0; i < NCFGCOLOURS; i++) {
    inst->cols[ww[i]] = 0x10000000 
      | ((inst->cfg.colours[i][0] & 0xff) << 16)
      | ((inst->cfg.colours[i][1] & 0xff) << 8)
      | (inst->cfg.colours[i][2] & 0xff);
  }

  for (i = 0; i < NEXTCOLOURS; i++) {
    if (i < 216) {
      int r = i / 36, g = (i / 6) % 6, b = i % 6;
      inst->cols[i+16] = 0x10000000
        | (((r ? r * 40 + 55 : 0) & 0xff) << 16)
        | (((g ? g * 40 + 55 : 0) & 0xff) << 8)
        | ((b ? b * 40 + 55 : 0) & 0xff);
    } else {
      int shade = i - 216;
      shade = (shade * 10 + 8) & 0xff;
      inst->cols[i+16] = 0x10000000 | (shade << 16) | (shade << 8) | shade;
    }
  }

  /* Since Default Background may have changed, ensure that space
   * between text area and window border is refreshed. */
  if (inst->window) {
    draw_backing_rect(inst);
    repaint(frontend);
  }
}

static void incr_rbar(W pid, W* val, int offset, int up)
{
  if (up) {
    if (val[0] - offset < val[2]){
      offset = val[0] - val[2];
    }
    val[0] -= offset;
    val[1] -= offset;
  }
  else {
    if (val[1] + offset > val[3]){
      offset = val[3] - val[1];
    }
    val[0] += offset;
    val[1] += offset;
  }
  cset_val(pid, 2, val);
}

void on_scrollbar(WEVENT *ev, void *frontend)
{
  W pid;
  W val[4];
  W act;
  int cont;
  struct gui_data *inst = (struct gui_data *)frontend;

  wget_bar(inst->window->wid, &pid, NULL, NULL);
  cget_val(pid, 4, &val);

  if (val[0] == val[2] && val[1] == val[3]) {
    return;
  }

  for (cont = TRUE; cont; ) {
    act = cact_par(pid, ev);
    cget_val(pid, 2, val);
    switch (act){
        case 0x6000: /* up(press) */
          incr_rbar(pid, val, 1, TRUE);
          break;
        case 0x6001: /* down(press) */
          incr_rbar(pid, val, 1, FALSE);
          break;
        case 0x5004: /* up(dck) */
          incr_rbar(pid, val, 3, TRUE);
          cont = FALSE;
          break;
        case 0x5005: /* down(dck) */
          incr_rbar(pid, val, 3, FALSE);
          cont = FALSE;
          break;
        case 0x5008: /* drag(up) */
        case 0x5009: /* drag(down) */
          cont = FALSE;
          break;
        default:
          cont = FALSE;
      }
  }

  set_sbar(frontend, val[3], val[0], val[1]-val[0]);
  term_scroll(inst->term, 1, val[0]);
}

void on_mousewheel(WEVENT *ev, void *frontend, int diff)
{
  W pid, wid;
  W val[4], prv;
  struct gui_data *inst = (struct gui_data *)frontend;

  wget_bar(inst->window->wid, &pid, NULL, NULL);
  cget_val(pid, 4, &val);
  prv = val[0];

  if (val[0] == val[2] && val[1] == val[3]) {
    return;
  }

  if (diff >= 0) {
    incr_rbar(pid, val, diff, FALSE);
  }
  else {
    incr_rbar(pid, val, -diff, TRUE);
  }

  set_sbar(frontend, val[3], val[0], val[1]-val[0]);
  term_scroll(inst->term, 1, val[0]);
}

int on_buttondown(WEVENT *ev, void *frontend)
{
  int cont = TRUE;
  struct gui_data *inst = (struct gui_data *)frontend;
  W clt;
  int w, h;
  RECT rsz;
  W err;

  if (ev->s.cmd != W_WORK && ev->s.cmd != W_RBAR) {
    clt = wchk_dck(ev->s.time);
  }

  switch (ev->s.cmd) {
      case W_WORK:
        break;
      case W_LTHD:
      case W_LBHD:
      case W_RTHD:
      case W_RBHD:
        if (clt == W_DCLICK){
          W full;
          wget_sts(inst->window->wid, &(inst->rsave), NULL);
          full = wchg_wnd(ev->s.wid, &rsz, W_MOVE);
          if (full >= 0) {
            w = (rsz.c.right - rsz.c.left - 2 * inst->cfg.window_border) 
              / inst->font_width;
            h = (rsz.c.bottom - rsz.c.top - 2 * inst->cfg.window_border) 
              / inst->font_height;
            inst->wfull = (full == 1);
            request_resize(frontend, w, h);
          }
        }
        else if (clt == W_PRESS || clt == W_QPRESS) {
          inst->window_moving = TRUE;
          err = wrsz_drg(ev, NULL, &rsz);
          inst->window_moving = FALSE;
          if (err >= 0) {
            inst->wfull = 0;
            if (err >= 1) {
              on_redisp(ev, frontend);
            }
          }
          w = (rsz.c.right - rsz.c.left - 2 * inst->cfg.window_border)
            / inst->font_width;
          h = (rsz.c.bottom - rsz.c.top - 2 * inst->cfg.window_border) 
            / inst->font_height;
          request_resize(frontend, w, h);
        }
        break;
      case W_PICT:
        if (clt == W_DCLICK){
          cont = FALSE;
          break;
        }
        else if (clt == W_CLICK){
          break; /* no-operation */
        }
        /* FALLTHROUGH */
      case W_TITL:
      case W_FRAM:
        if (clt == W_PRESS || clt == W_QPRESS){
          inst->window_moving = TRUE;
          err = wmov_drg(ev, NULL);
          inst->window_moving = FALSE;
          if (err >= 0) {
            inst->wfull = 0;
            if (err >= 1){
              on_redisp(ev, frontend);
            }
          }
        }
        break;
      case W_RBAR:
        on_scrollbar(ev, frontend);
        /* TODO */
        break;
      default:
        break;
  }

  return cont;
}

void on_idle(WEVENT *ev, void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;

  if (ev->s.cmd == W_WORK) {
    PTRSTL pstl = 
      ((ev->s.stat & (ES_LSHFT | ES_RSHFT)) != 0) 
      ? PS_MODIFY : PS_SELECT;
    gset_ptr(pstl, NULL, (COLOR)-1, (COLOR)-1);
  }

  /* TODO: motion_event */
}

void repaint(void* frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  GID gid;
  RECT wr;
  RECT r;

  if (inst->window->wid < 0 || inst->pixmap == NULL) {
    return;
  }
  
  gid = wget_gid(inst->window->wid);

  /*
   * top-left corner is always (0,0)
   * gset_vis, gset_fra is needed in Chokanji V systems.
   */
  wr.c.left = wr.c.top = 0;
  wset_wrk(inst->window->wid, &wr);
  gset_vis(gid, wr);
  gset_fra(gid, wr);

  r = inst->pixmap->bounds;
  grst_bmp(gid, NULL, inst->pixmap, &r, NULL, G_STORE);
}

void on_redisp(WEVENT *ev, void *frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;

  if (inst->window->wid < 0) {
    return;
  }
  do {
    if (wsta_dsp(inst->window->wid, NULL, NULL) == 0)
      break;
    repaint(frontend);
  } while (wend_dsp(inst->window->wid) > 0);
}

void on_focus(WEVENT *ev, void *frontend, int has_focus)
{
  struct gui_data *inst = (struct gui_data *)frontend;
  term_set_focus(inst->term, has_focus);
  term_update(inst->term);
}

void on_message(WEVENT *ev, void* frontend)
{
  W err;
  struct fd_input_message msg;
  
  err = rcv_msg(MM_ALL, (MESSAGE*)&msg, sizeof(msg), NOWAIT|CLR);
  if (err < 0) {
    clr_msg(MM_ALL, MM_ALL);
    return;
  }

  if (err != 0 /* other process has sent this message */
      || msg.msg_type != MS_TYPE7 
      || msg.msg_size != sizeof(msg) - sizeof(W)*2) {
    return;
  }

  /* process fd_input events */
  if ((msg.rwx & 4) != 0)
    select_result(msg.fd->fd, 4);
  if ((msg.rwx & 1) != 0)
    select_result(msg.fd->fd, 1);
  if ((msg.rwx & 2) != 0)
    select_result(msg.fd->fd, 2);

  /* notify completion to the selecter-task */
  wup_tsk(msg.fd->tskid);
}

void on_keydown(WEVENT *ev, void* frontend)
{
  struct gui_data *inst = (struct gui_data *)frontend;

  char output[32];
  wchar_t ucsoutput[2];
  TC tmp[2];
  int ucsval, start, end, special, use_ucsoutput;
  unsigned int keysym;

  /* By default, nothing is generated. */
  end = start = 0;
  special = use_ucsoutput = FALSE;

  /*
   * NYI: Compose key (!!! requires Unicode faff before even trying)
   */

  /*
   * Shift-PgUp and Shift-PgDn don't even generate keystrokes
   * at all.
   */
  if (ev->e.data.key.code == KC_PGUP && (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))) {
    term_scroll(inst->term, 0, -inst->cfg.height/2);
    return;
  }
  if (ev->e.data.key.code == KC_PGUP && (ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD))) {
    term_scroll(inst->term, 0, -1);
    return;
  }
  if (ev->e.data.key.code == KC_PGDN && (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))) {
    term_scroll(inst->term, 0, +inst->cfg.height/2);
    return;
  }
  if (ev->e.data.key.code == KC_PGDN && (ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD))) {
    term_scroll(inst->term, 0, +1);
    return;
  }

  /*
   * Neither does Shift-Ins.
   */
  if (ev->e.data.key.code == INS_CODE && (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))) {
    request_paste(inst);
    return;
  }

  special = FALSE;
  use_ucsoutput = FALSE;

  /* ALT+things gives leading Escape. */
  output[0] = '\033';
  sprintf(output+1, "%C", ev->e.data.key.code);

  if ((ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD)) != 0) {
    if (output[1] >= 'A' && output[1] <= 'Z') {
      output[1] -= ('A'-1);
    }
    else if (output[1] >= 'a' && output[1] <= 'z') {
      output[1] -= ('a'-1);
    }
    else {
      /* BTRON TODO */
      return;
    }
  }
  keysym = output[1];

  if ((ucsval = keysym_to_unicode(keysym)) >= 0) {
    ucsoutput[0] = '\033';
    ucsoutput[1] = ucsval;
    use_ucsoutput = TRUE;
    end = 2;
  } else {
    output[31] = '\0';
    end = strlen(output);
  }
  if (ev->e.stat & (ES_EXT|ES_LEXT|ES_TEXT)) {
    start = 0;
    if (end == 1) end = 0;
  } else
    start = 1;

  /* We handle Return ourselves, because it needs to be flagged as
   * special to ldisc. */
  if (ev->e.data.key.code == KC_CR || ev->e.data.key.code == KC_NL) { /* TODO: not sure */
    output[1] = '\015';
    use_ucsoutput = FALSE;
    end = 2;
    special = TRUE;
  }

  /* Control-2, Control-Space and Control-@ are NUL */
  if ((ev->e.data.key.code == KC_SPACE || ev->e.data.key.code == TK_2 ||
       ev->e.data.key.code == TK_ATMK) &&
      (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT)) == 0 &&
      (ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD)) != 0) {
    output[1] = '\0';
    use_ucsoutput = FALSE;
    end = 2;
  }

  /* Control-Shift-Space is 160 (ISO8859 nonbreaking space) */
  if (ev->e.data.key.code == KC_SPACE &&
      (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT)) != 0 &&
      (ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD)) != 0) {
    output[1] = '\240';
    use_ucsoutput = FALSE;
    end = 2;
  }

  /* We don't let GTK tell us what Backspace is! We know better. */
  if (ev->e.data.key.code == KC_BS &&
      !(ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))) {
    output[1] = inst->cfg.bksp_is_delete ? '\x7F' : '\x08';
    use_ucsoutput = FALSE;
    end = 2;
    special = TRUE;
  }
  /* For Shift Backspace, do opposite of what is configured. */
  if (ev->e.data.key.code == KC_BS &&
      (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))) {
    output[1] = inst->cfg.bksp_is_delete ? '\x08' : '\x7F';
    use_ucsoutput = FALSE;
    end = 2;
    special = TRUE;
  }

  /* Shift-Tab is ESC [ Z */
  if (ev->e.data.key.code == KC_TAB 
      && (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))) {
    end = 1 + sprintf(output+1, "\033[Z");
    use_ucsoutput = FALSE;
  }

  /*
   * NetHack keypad mode.
   */
  if (inst->cfg.nethack_keypad) {
    char *keys = NULL;
    switch (ev->e.data.key.code) {
        case TK_1: case KC_END: keys = "bB\002"; break;
        case TK_2: case KC_CC_D: keys = "jJ\012"; break;
        case TK_3: case KC_PGDN: keys = "nN\016"; break;
        case TK_4: case KC_CC_L: keys = "hH\010"; break;
        case TK_5: /* case GDK_KP_Begin: */ keys = "..."; break;
        case TK_6: case KC_CC_R: keys = "lL\014"; break;
        case TK_7: case KC_HOME: keys = "yY\031"; break;
        case TK_8: case KC_CC_U: keys = "kK\013"; break;
        case TK_9: case KC_PGUP: keys = "uU\025"; break;
    }
    if (keys) {
      end = 2;
      if (ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD))
        output[1] = keys[2];
      else if (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))
        output[1] = keys[1];
      else
        output[1] = keys[0];
      use_ucsoutput = FALSE;
      goto done;
    }
  }

#if 0 /* BTRON */
  /*
   * Application keypad mode.
   */
  if (inst->term->app_keypad_keys && !inst->cfg.no_applic_k) {
    int xkey = 0;
    switch (ev->e.data.key.code) {
//	      case GDK_Num_Lock: xkey = 'P'; break;
        case TK_ASTR: xkey = 'Q'; break;
        case TK_SLSH: xkey = 'R'; break;
        case TK_MINS: xkey = 'S'; break;
          /*
           * Keypad + is tricky. It covers a space that would
           * be taken up on the VT100 by _two_ keys; so we
           * let Shift select between the two. Worse still,
           * in xterm function key mode we change which two...
           */
        case TK_PLUS:
          if (inst->cfg.funky_type == FUNKY_XTERM) {
            if (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))
              xkey = 'l';
            else
              xkey = 'k';
          } else if (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT))
            xkey = 'm';
          else
            xkey = 'l';
          break;
        case KC_NL: case KC_CR: xkey = 'M'; break;
        case TK_0: case INS_CODE: xkey = 'p'; break;
        case TK_1: /*case GDK_KP_End:*/ xkey = 'q'; break;
        case TK_2: case KC_CC_D: xkey = 'r'; break;
        case TK_3: case KC_PGDN: xkey = 's'; break;
        case TK_4: case KC_CC_L: xkey = 't'; break;
        case TK_5: /*case GDK_KP_Begin:*/ xkey = 'u'; break;
        case TK_6: case KC_CC_R: xkey = 'v'; break;
        case TK_7: case KC_HOME: xkey = 'w'; break;
        case TK_8: case KC_CC_U: xkey = 'x'; break;
        case TK_9: case KC_PGUP: xkey = 'y'; break;
//        case GDK_KP_Decimal: case GDK_KP_Delete: xkey = 'n'; break;
    }
    if (xkey) {
      if (inst->term->vt52_mode) {
        if (xkey >= 'P' && xkey <= 'S')
          end = 1 + sprintf(output+1, "\033%c", xkey);
        else
          end = 1 + sprintf(output+1, "\033?%c", xkey);
      } else
        end = 1 + sprintf(output+1, "\033O%c", xkey);
      use_ucsoutput = FALSE;
      goto done;
    }
  }
#endif

  /*
   * Next, all the keys that do tilde codes. (ESC '[' nn '~',
   * for integer decimal nn.)
   *
   * We also deal with the weird ones here. Linux VCs replace F1
   * to F5 by ESC [ [ A to ESC [ [ E. rxvt doesn't do _that_, but
   * does replace Home and End (1~ and 4~) by ESC [ H and ESC O w
   * respectively.
   */
  {
    int code = 0;
    switch (ev->e.data.key.code) {
        case KC_PF1:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 23 : 11);
          break;
        case KC_PF2:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 24 : 12);
          break;
        case KC_PF3:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 25 : 13);
          break;
        case KC_PF4:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 26 : 14);
          break;
        case KC_PF5:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 28 : 15);
          break;
        case KC_PF6:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 29 : 17);
          break;
        case KC_PF7:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 31 : 18);
          break;
        case KC_PF8:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 32 : 19);
          break;
        case KC_PF9:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 33 : 20);
          break;
        case KC_PF10:
          code = (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT) ? 34 : 21);
          break;
        case KC_PF11:
          code = 23;
          break;
        case KC_PF12:
          code = 24;
          break;
    }
    if (!(ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD))) 
      switch (ev->e.data.key.code) {
          case KC_HOME:
            code = 1;
            break;
          case INS_CODE:
            code = 2;
            break;
          case KC_DEL:
            code = 3;
            break;
          case KC_END:
            code = 4;
            break;
          case KC_PGUP:
            code = 5;
            break;
          case KC_PGDN:
            code = 6;
            break;
      }
    /* Reorder edit keys to physical order */
    if (inst->cfg.funky_type == FUNKY_VT400 && code <= 6)
      code = "\0\2\1\4\5\3\6"[code];

    if (inst->term->vt52_mode && code > 0 && code <= 6) {
      end = 1 + sprintf(output+1, "\x1B%c", " HLMEIG"[code]);
      use_ucsoutput = FALSE;
      goto done;
    }

    if (inst->cfg.funky_type == FUNKY_SCO &&     /* SCO function keys */
        code >= 11 && code <= 34) {
      char codes[] = "MNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz@[\\]^_`{";
      int index = 0;
      switch (ev->e.data.key.code) {
          case KC_PF1: index = 0; break;
          case KC_PF2: index = 1; break;
          case KC_PF3: index = 2; break;
          case KC_PF4: index = 3; break;
          case KC_PF5: index = 4; break;
          case KC_PF6: index = 5; break;
          case KC_PF7: index = 6; break;
          case KC_PF8: index = 7; break;
          case KC_PF9: index = 8; break;
          case KC_PF10: index = 9; break;
          case KC_PF11: index = 10; break;
          case KC_PF12: index = 11; break;
      }
      if (ev->e.stat & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT)) index += 12;
      if (ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD)) index += 24;
      end = 1 + sprintf(output+1, "\x1B[%c", codes[index]);
      use_ucsoutput = FALSE;
      goto done;
    }
    if (inst->cfg.funky_type == FUNKY_SCO &&     /* SCO small keypad */
        code >= 1 && code <= 6) {
      char codes[] = "HL.FIG";
      if (code == 3) {
        output[1] = '\x7F';
        end = 2;
      } else {
        end = 1 + sprintf(output+1, "\x1B[%c", codes[code-1]);
      }
      use_ucsoutput = FALSE;
      goto done;
    }
    if ((inst->term->vt52_mode || inst->cfg.funky_type == FUNKY_VT100P) &&
        code >= 11 && code <= 24) {
      int offt = 0;
      if (code > 15)
        offt++;
      if (code > 21)
        offt++;
      if (inst->term->vt52_mode)
        end = 1 + sprintf(output+1,
                          "\x1B%c", code + 'P' - 11 - offt);
      else
        end = 1 + sprintf(output+1,
                          "\x1BO%c", code + 'P' - 11 - offt);
      use_ucsoutput = FALSE;
      goto done;
    }
    if (inst->cfg.funky_type == FUNKY_LINUX && code >= 11 && code <= 15) {
      end = 1 + sprintf(output+1, "\x1B[[%c", code + 'A' - 11);
      use_ucsoutput = FALSE;
      goto done;
    }
    if (inst->cfg.funky_type == FUNKY_XTERM && code >= 11 && code <= 14) {
      if (inst->term->vt52_mode)
        end = 1 + sprintf(output+1, "\x1B%c", code + 'P' - 11);
      else
        end = 1 + sprintf(output+1, "\x1BO%c", code + 'P' - 11);
      use_ucsoutput = FALSE;
      goto done;
    }
    if (inst->cfg.rxvt_homeend && (code == 1 || code == 4)) {
      end = 1 + sprintf(output+1, code == 1 ? "\x1B[H" : "\x1BOw");
      use_ucsoutput = FALSE;
      goto done;
    }
    if (code) {
      end = 1 + sprintf(output+1, "\x1B[%d~", code);
      use_ucsoutput = FALSE;
      goto done;
    }
  }

  /*
   * Cursor keys. (This includes the numberpad cursor keys,
   * if we haven't already done them due to app keypad mode.)
   * 
   * Here we also process un-numlocked un-appkeypadded KP5,
   * which sends ESC [ G.
   */
  {
    int xkey = 0;
    switch (ev->e.data.key.code) {
        case KC_CC_U: xkey = 'A'; break;
        case KC_CC_D: xkey = 'B'; break;
        case KC_CC_R: xkey = 'C'; break;
        case KC_CC_L: xkey = 'D'; break;
//        case GDK_Begin: case GDK_KP_Begin: xkey = 'G'; break;
    }
    if (xkey) {
      /*
       * The arrow keys normally do ESC [ A and so on. In
       * app cursor keys mode they do ESC O A instead.
       * Ctrl toggles the two modes.
       */
      if (inst->term->vt52_mode) {
        end = 1 + sprintf(output+1, "\033%c", xkey);
      } else if (!inst->term->app_cursor_keys ^
                 !(ev->e.stat & (ES_CMD|ES_LCMD|ES_TCMD))) {
        end = 1 + sprintf(output+1, "\033O%c", xkey);
      } else {		    
        end = 1 + sprintf(output+1, "\033[%c", xkey);
      }
      use_ucsoutput = FALSE;
      goto done;
    }
  }
  goto done;

done:

  if (end-start > 0) {
#ifdef KEY_DEBUGGING
    int i;
    printf("generating sequence:");
    for (i = start; i < end; i++)
      printf(" %02x", (unsigned char) output[i]);
    printf("\n");
#endif

    if (special) {
      /*
       * For special control characters, the character set
       * should never matter.
       */
      output[end] = '\0';	       /* NUL-terminate */
      if (inst->ldisc)
        ldisc_send(inst->ldisc, output+start, -2, 1);
    } else if (!inst->direct_to_font) {
      if (!use_ucsoutput) {
        /*
         * The stuff we've just generated is assumed to be
         * ISO-8859-1! This sounds insane, but `man
         * XLookupString' agrees: strings of this type
         * returned from the X server are hardcoded to
         * 8859-1. Strictly speaking we should be doing
         * this using some sort of GtkIMContext, which (if
         * we're lucky) would give us our data directly in
         * Unicode; but that's not supported in GTK 1.2 as
         * far as I can tell, and it's poorly documented
         * even in 2.0, so it'll have to wait.
         */
        if (inst->ldisc)
          lpage_send(inst->ldisc, CS_ISO8859_1, output+start,
                     end-start, 1);
      } else {
        /*
         * We generated our own Unicode key data from the
         * keysym, so use that instead.
         */
        if (inst->ldisc)
          luni_send(inst->ldisc, ucsoutput+start, end-start, 1);
      }
    } else {
      /*
       * In direct-to-font mode, we just send the string
       * exactly as we received it.
       */
      if (inst->ldisc)
        ldisc_send(inst->ldisc, output+start, end-start, 1);
    }

//    show_mouseptr(inst, 0);
    term_seen_key_event(inst->term);
  }
}

#if 0 /* BTRON */
gint button_event(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
    struct gui_data *inst = (struct gui_data *)data;
    int shift, ctrl, alt, x, y, button, act;

    /* Remember the timestamp. */
    inst->input_event_time = event->time;

    show_mouseptr(inst, 1);

    if (event->button == 4 && event->type == GDK_BUTTON_PRESS) {
	term_scroll(inst->term, 0, -5);
	return TRUE;
    }
    if (event->button == 5 && event->type == GDK_BUTTON_PRESS) {
	term_scroll(inst->term, 0, +5);
	return TRUE;
    }

    shift = event->state & (ES_LLSHFT|ES_LRSHFT|ES_TLSHFT|ES_TRSHFT);
    ctrl = event->state & GDK_CONTROL_MASK;
    alt = event->state & GDK_MOD1_MASK;

    if (event->button == 3 && ctrl) {
	gtk_menu_popup(GTK_MENU(inst->menu), NULL, NULL, NULL, NULL,
		       event->button, event->time);
	return TRUE;
    }

    if (event->button == 1)
	button = MBT_LEFT;
    else if (event->button == 2)
	button = MBT_MIDDLE;
    else if (event->button == 3)
	button = MBT_RIGHT;
    else
	return FALSE;		       /* don't even know what button! */

    switch (event->type) {
      case GDK_BUTTON_PRESS: act = MA_CLICK; break;
      case GDK_BUTTON_RELEASE: act = MA_RELEASE; break;
      case GDK_2BUTTON_PRESS: act = MA_2CLK; break;
      case GDK_3BUTTON_PRESS: act = MA_3CLK; break;
      default: return FALSE;	       /* don't know this event type */
    }

    if (send_raw_mouse && !(inst->cfg.mouse_override && shift) &&
	act != MA_CLICK && act != MA_RELEASE)
	return TRUE;		       /* we ignore these in raw mouse mode */

    x = (event->x - inst->cfg.window_border) / inst->font_width;
    y = (event->y - inst->cfg.window_border) / inst->font_height;

    term_mouse(inst->term, button, translate_button(button), act,
	       x, y, shift, ctrl, alt);

    return TRUE;
}
#endif

static void start_backend(struct gui_data *inst)
{
  extern Backend *select_backend(Config *cfg);
  char *realhost;
  const char *error;

  inst->back = select_backend(&inst->cfg);

  error = inst->back->init((void *)inst, &inst->backhandle,
                           &inst->cfg, inst->cfg.host, inst->cfg.port,
                           &realhost, inst->cfg.tcp_nodelay,
                           inst->cfg.tcp_keepalives);

  if (error) {
    char *msg = dupprintf("Unable to open connection to %s:\n%s",
                          inst->cfg.host, error);
    inst->exited = TRUE;
    fatal_message_box(inst->window, msg);
    sfree(msg);
    /* TODO: add appropreate finalization */
    oend_prc(((M_EXECREQ*)(inst->msg))->vid, inst->fusen, 0);
    ext_prc(0);
  }

  if (inst->cfg.wintitle[0]) {
    set_title(inst, inst->cfg.wintitle);
    set_icon(inst, inst->cfg.wintitle);
  } else {
    char *title = make_default_wintitle(realhost);
    set_title(inst, title);
    set_icon(inst, title);
    sfree(title);
  }
  sfree(realhost);

//  inst->back->provide_logctx(inst->backhandle, inst->logctx);

  term_provide_resize_fn(inst->term, inst->back->size, inst->backhandle);

  inst->ldisc =
    ldisc_create(&inst->cfg, inst->term, inst->back, inst->backhandle,
                 inst);

//  gtk_widget_hide(inst->restartitem);
}

W pt_main(MESSAGE *msg, LINK *lnk, putty_fusen* fusen)
{
  struct gui_data *inst;
  RECT wrect;
  PAT wbgpat;
  TC wtitle[] = {TCS_PUTTY};
  TC fname_conf[] = {TCS_FNAME_CONFIG};
  int cont;
  WEVENT wev0;
  LINK conf;
  putty_fusen newfusen;
  WDSTAT stat;
  UW attr;

  dopn_dat(NULL);
  getscreen();
  initstdpnl();

  /* change working directory to '/SYS/$$PROGRAM.BOX/SSHCLIENT/config' */
  prc_inf(0, PI_LINK, &conf, sizeof(LINK));
  get_lnk(fname_conf, &conf, F_BASED);
  chg_wrk(&conf);

  /*
   * Create an instance structure and initialise to zeroes
   */
  inst = snew(struct gui_data);
  memset(inst, 0, sizeof(*inst));
  inst->busy_status = BUSY_NOT;

  if (fusen != NULL) {
    memcpy(&newfusen, fusen, sizeof(newfusen));
  }
  inst->msg = msg;
  inst->fusen = fusen;

  GetStdCSPEC(&(inst->cspec), SCREEN.planes, SCREEN.pixbits);

  do_defaults(lnk, &inst->cfg);
  cmdline_run_saved(&inst->cfg);

  /* originally, this was setup_fonts_ucs(inst); */
  inst->font_width = 8; /* TODO */
  inst->font_height = 16; /* TODO */
  inst->direct_to_font = init_ucs(&inst->ucsdata, inst->cfg.line_codepage,
                                  inst->cfg.utf8_override, CS_ISO8859_1, /* TODO */
                                  inst->cfg.vtmode);
  inst->cfg.window_border = 1;
  inst->mouse_x = -1;
  inst->mouse_y = -1;

#if 0 /* BTRON TODO */
  init_cutbuffers();
#endif

  /*
   * Set up the colour map.
   */
  palette_reset(inst);

  /* singleton; no need for sfree */
  inst->window = snew(Window);
  memset(inst->window, 0, sizeof(*(inst->window)));

  attr = MAIN_WINDOW_ATTR;
  if (fusen != NULL) {
    wrect.p.lefttop = fusen->wr.p.lefttop;
    inst->wfull = fusen->is_full;
    if (inst->wfull) {
      attr |= WA_FULL;
    }
  }
  else {
    wrect.c.left = 50;
    wrect.c.right = 50;
    inst->wfull = 0;
  }
  wrect.c.right = wrect.c.left + inst->font_width * inst->cfg.width + 2 * inst->cfg.window_border;
  wrect.c.bottom = wrect.c.top + inst->font_height * inst->cfg.height + 2 * inst->cfg.window_border;
  inst->rsave.r = wrect;

  wbgpat.spat.kind = 0;
  wbgpat.spat.hsize = wbgpat.spat.vsize = 1;
  wbgpat.spat.fgcol = wbgpat.spat.bgcol = inst->cols[258];
  wbgpat.spat.mask = FILL100;

  inst->window->wid = wopn_wnd(attr, 0, &wrect, NULL, 17, wtitle, &wbgpat, NULL);
  if (inst->window->wid < 0) {
    /* TODO: add appropreate finalization */
    oend_prc(((M_EXECREQ*)msg)->vid, inst->fusen, 0);
    ext_prc(1);
  }

  osta_prc(((M_EXECREQ*)msg)->vid, inst->window->wid);

  inst->term = term_init(&inst->cfg, &inst->ucsdata, inst);
  uxsel_init();
  term_size(inst->term, inst->cfg.height, inst->cfg.width, inst->cfg.savelines);
  start_backend(inst);
  ldisc_send(inst->ldisc, NULL, 0, 0); /* cause ldisc to notice changes */
  request_resize(inst, inst->cfg.width, inst->cfg.height);

  inst->exited = FALSE;

  /* event loop */
  for (cont = TRUE; cont && !inst->exited; ) {
    wget_evt(&wev0, WAIT);

    switch (wev0.s.type) {
        case EV_NULL:
        case EV_NOMSG:
          on_idle(&wev0, inst);
          break;
        case EV_REQUEST:
          switch (wev0.g.cmd) {
              case W_REDISP:
                on_redisp(&wev0, inst);
                break;
              case W_PASTE:
                wrsp_evt(&wev0, 1); /* NACK */
                break;
              case W_DELETE:
              case W_FINISH:
                wrsp_evt(&wev0, 0); /* ACK */
                cont = FALSE;
                break;
          }
          break;
        case EV_RSWITCH:
          on_redisp(&wev0, inst);
          /* FALLTHROUGH */
        case EV_SWITCH:
          on_focus(&wev0, inst, TRUE);
          /* FALLTHROUGH */
        case EV_BUTDWN:
          cont = on_buttondown(&wev0, inst);
          break;
        case EV_INACT:
          on_focus(&wev0, inst, FALSE);
          break;
        case EV_DEVICE:
          oprc_dev(&(wev0.e), NULL, 0);
          break;
        case EV_MSG:
          on_message(&wev0, inst);
          break;
        case EV_MENU:
//          on_menu(&wev0, inst);
          break;
        case EV_AUTKEY:
          switch (wev0.e.data.key.code) {
              case KC_SS_D:
              case KC_SS_U:
                on_mousewheel(&wev0, inst, wev0.e.data.key.keytop - 0x8000);
                continue;
              default:
                break;
          }
          /* FALLTHROUGH */
        case EV_KEYDWN:
          on_keydown(&wev0, inst);
          break;
    }
  }

  /* oend_prc is done inside this `put_fsn_data' */
  wget_sts(inst->window->wid, &stat, NULL);
  newfusen.dlen = sizeof(newfusen) - sizeof(UH);
  newfusen.is_full = inst->wfull ? 1 : 0;
  newfusen.wr = (newfusen.is_full ? inst->rsave.r : stat.r);
  put_fsn_data((M_EXECREQ*)msg, &newfusen, fusen, True);

  if (inst->window->wid >= 0) {
    wcls_wnd(inst->window->wid, NOCLR /*CLR*/);
    inst->window->wid = -1;
  }

  /* Without this, other windows somehow seem tainted (even with CLR) */
  wreq_dsp(0);

  /* TODO: add extra finalization here */
  return 0;
}

/*
 * TODO: Unconnected events are
 *         - motion_event
 *         - button_event
 *         - selection_received
 *         - selection_get
 */
