/* BTRON File I/O */

#include <basic.h>
#include <btron/file.h>
#include <btron/vobj.h>
#include <bassert.h>
#include <keycode.h>
#include <tcode.h>

#define GLOBAL_DEFINE_BTRON_FILEIO
#include "fileio.h"
#include "putty.h"
#include "puttymem.h"

/* So far, we only support "w", "wb", "r", "rb". */
FILE* f_open_rec(struct Filename filename, const char *mode, int is_private, int rectype)
{
  FILE *fp = NULL;
  FILE_open_type open_type = FILE_OPEN_UNKNOWN;
  W err;
  W r_size;

  if (strcmp(mode, "r") == 0 || strcmp(mode, "rb") == 0) {
    open_type = FILE_OPEN_READ;
  }
  else if (strcmp(mode, "w") == 0 || strcmp(mode, "wb") == 0) {
    open_type = FILE_OPEN_WRITE;
  }
  else {
    return NULL;
  }

  fp = snew(FILE);
  fp->fd = -1;
  fp->open_type = open_type;
  fp->current_record_pos = 0;
  fp->buffer_start_index = 0;
  fp->buffer_end_index = 0;
  fp->recnum = -1;

  switch (open_type) {
      case FILE_OPEN_READ:
        if (filename.tag == FILENAME_LINK) {
          fp->fd = opn_fil(&(filename.link), F_READ, NULL);
          if (fp->fd < 0) {
            goto err_ret;
          }
        }
        else if (filename.tag == FILENAME_LINKNAME) {
          LINK lnk = filename.link;
          err = get_lnk(filename.path, &lnk, F_BASED);
          if (err < 0) {
            goto err_ret;
          }
          fp->fd = opn_fil(&lnk, F_READ, NULL);
          if (fp->fd < 0) {
            goto err_ret;
          }
        }
        else {
          goto err_ret;
        }
        err = fnd_rec(fp->fd, F_TOPEND, 1 << rectype, 0, NULL);
        if (err < 0) {
          goto err_ret;
        }
        err = rea_rec(fp->fd, 0, NULL, 0, &r_size, NULL);
        if (err < 0) {
          goto err_ret;
        }
        fp->record_size = r_size;
        break;
      case FILE_OPEN_WRITE:
        if (filename.tag == FILENAME_LINK) {
          fp->fd = opn_fil(&(filename.link), F_WRITE, NULL);
          if (fp->fd < 0) {
            goto err_ret;
          }
        }
        else if (filename.tag == FILENAME_LINKNAME) {
          LINK lnk = filename.link;
          err = get_lnk(filename.path, &lnk, F_BASED);
          if (err < 0) {
            fp->fd = cre_fil(&(filename.link), filename.path, NULL, 2, F_FIX);
            if (fp->fd < 0) {
              goto err_ret;
            }
          }
          else {
            fp->fd = opn_fil(&lnk, F_WRITE, NULL);
            if (fp->fd < 0) {
              goto err_ret;
            }
          }
        }
        else {
          goto err_ret;
        }
        err = fnd_rec(fp->fd, F_TOPEND, 1 << rectype, 0, NULL);
        if (err < 0) {
          err = apd_rec(fp->fd, NULL, 0, rectype, 0, 0);
          if (err < 0) {
            goto err_ret;
          }
          err = see_rec(fp->fd, -1, -1, NULL);
          if (err < 0) {
            goto err_ret;
          }            
        }
        fp->record_size = 0; /* TODO: trc_rec to this size when closing */
        break;
      default:
        goto err_ret;
  }

  return fp;

err_ret:
  if (fp != NULL) {
    if (fp->fd > 0) {
      cls_fil(fp->fd);
    }
    sfree(fp);
  }
  return NULL;
}

int fflush(FILE *fp)
{
  W err, len;

  if (fp == NULL || fp->fd < 0) {
    return -1;
  }
  if (fp->open_type != FILE_OPEN_WRITE) {
    return 0;
  }

  len = fp->buffer_end_index - fp->buffer_start_index;
  if (len > 0) {
    err = wri_rec(fp->fd, fp->record_size, &(fp->buf[fp->buffer_start_index]), 
                  len, NULL, NULL, 0);
    if (err < 0) {
      return -1;
    }
    fp->record_size += len;
  }

  fp->buffer_start_index = 0;
  fp->buffer_end_index = 0;

  return 0;
}

int fclose(FILE *fp)
{
  if (fp == NULL) {
    return -1;
  }

  if (fp->open_type == FILE_OPEN_WRITE) {
    fflush(fp);
    trc_rec(fp->fd, fp->record_size);
  }

  if (fp->fd >= 0) {
    cls_fil(fp->fd);
  }

  sfree(fp);

  return 0;
}

static size_t fskip(FILE *fp, int bytes)
{
  int rsize = bytes;
  int len = fp->buffer_end_index - fp->buffer_start_index;

  if (rsize < len) {
    fp->buffer_start_index += rsize;
    return bytes;
  }
  rsize -= len;
  fp->current_record_pos += fp->buffer_end_index;
  fp->buffer_start_index = 0;
  fp->buffer_end_index = 0;

  if (fp->current_record_pos + rsize <= fp->record_size) {
    fp->current_record_pos += rsize;
    return bytes;
  }
  else {
    rsize -= (fp->record_size - fp->current_record_pos);
    fp->current_record_pos = fp->record_size;
    return bytes - rsize;
  }
}

size_t fread(B *ptr, size_t size, size_t nmemb, FILE *fp)
{
  int rsize = size * nmemb;
  int len = fp->buffer_end_index - fp->buffer_start_index;
  int brsize;
  W err;

label_cont:

  if (len > 0 && rsize <= len) {
    memcpy(ptr, &(fp->buf[fp->buffer_start_index]), rsize);
    fp->buffer_start_index += rsize;
    if (rsize == len) {
      fp->current_record_pos += fp->buffer_end_index;
      fp->buffer_start_index = 0;
      fp->buffer_end_index = 0;
    }
    return nmemb;
  }
  else if (len > 0 && rsize > len) {
    memcpy(ptr, &(fp->buf[fp->buffer_start_index]), len);
    rsize -= len;
    ptr += len;
    fp->current_record_pos += fp->buffer_end_index;
    fp->buffer_start_index = 0;
    fp->buffer_end_index = 0;
  }

  assert(fp->buffer_start_index == fp->buffer_end_index);
  fp->current_record_pos += fp->buffer_start_index;
  fp->buffer_start_index = 0;
  fp->buffer_end_index = 0;

  if (fp->record_size == fp->current_record_pos) {
    return (size * nmemb - rsize) / size; /* EOF */
  }

  /* read remaining rsize-bytes */
  if (rsize >= BTRON_FILEIO_BUFSIZE 
      && fp->current_record_pos + rsize <= fp->record_size) {
    brsize = rsize;
    err = rea_rec(fp->fd, fp->current_record_pos, ptr, brsize, NULL, NULL);
    if (err < 0) {
      return (size * nmemb - rsize) / size;
    }
    fp->current_record_pos += brsize;
    return nmemb;
  }
  else if (rsize >= BTRON_FILEIO_BUFSIZE 
           && fp->current_record_pos + rsize > fp->record_size) {
    brsize = fp->record_size - fp->current_record_pos;
    err = rea_rec(fp->fd, fp->current_record_pos, ptr, brsize, NULL, NULL);
    if (err < 0) {
      return (size * nmemb - rsize) / size;
    }
    fp->current_record_pos += brsize;
    return (size * nmemb - rsize) / size;
  }

  /* read to buffer and continue */
  brsize = BTRON_FILEIO_BUFSIZE;
  if (brsize > fp->record_size - fp->current_record_pos) {
    brsize = fp->record_size - fp->current_record_pos;
  }
  err = rea_rec(fp->fd, fp->current_record_pos, fp->buf, brsize, NULL, NULL);
  if (err < 0) {
    return (size * nmemb - rsize) / size;
  }
  len = fp->buffer_end_index = brsize;

  goto label_cont;
}

int fgetc(FILE *fp)
{
  int err;
  unsigned char c;

  err = fread(&c, sizeof(c), 1, fp);
  if (err <= 0) {
    return EOF;
  }
  return c;
}

TC* fgets_tc(TC* s, int size, FILE *fp)
{
  int err;
  int i;
  TC c;
  UH tadlen;
  UW ltadlen;

  for (i = 0; i < size-1; i++) {
label_again:
    err = fread(&c, sizeof(c), 1, fp);
    if (err <= 0) {
      if (i != 0) {
        break;
      }
      return NULL;
    }
    if ((c & 0xff00) == 0xff00) {
      /* skip tad segment */
      err = fread(&tadlen, sizeof(tadlen), 1, fp);
      if (err <= 0) {
        if (i != 0) {
          break;
        }
        return NULL;
      }
      if (tadlen != 0xffff) {
        fskip(fp, tadlen);
      }
      else {
        err = fread(&ltadlen, sizeof(ltadlen), 1, fp);
        if (err <= 0) {
          if (i != 0) {
            break;
          }
          return NULL;
        }
        fskip(fp, ltadlen);
      }
#if 1 /* dirty hack to show `gppfile' that vobj was there */
      if ((c & 0xffff) == (0xff00 | TS_VOBJ)) {
        W recnum, err;
        see_rec(fp->fd, 0, 0, &recnum); /* save recnum */
        if (fp->recnum < 0) {
          err = fnd_rec(fp->fd, F_TOPEND, RM_LINK, 0, &(fp->recnum));
        }
        else {
          see_rec(fp->fd, fp->recnum, 1, NULL);
          err = fnd_rec(fp->fd, F_NFWD, RM_LINK, 0, &(fp->recnum));
        }
        if (err >= 0) {
          LINK link;
          B* bp = (B*)&link;
          TC hn[] = 
            {TK_0, TK_1, TK_2, TK_3, TK_4, TK_5, TK_6, TK_7,
             TK_8, TK_9, TK_A, TK_B, TK_C, TK_D, TK_E, TK_F};
          int j;

          err = rea_rec(fp->fd, 0, &link, sizeof(LINK), NULL, NULL);
          if (err >= 0) {
            for (j = 0; j < sizeof(LINK); j++) {
              if (i >= size-2) break;
              s[i] = hn[(bp[j] >> 4) & 0xf];
              s[i+1] = hn[bp[j] & 0xf];
              i += 2;
            }
          }
        }
        see_rec(fp->fd, recnum, 1, NULL); /* restore recnum */
      }
#endif
      goto label_again;
    }
    if (c == KC_NL || c == KC_CR) {
      s[i] = c;
      i++;
      break;
    }
    s[i] = c;
  }
  s[i] = TNULL;

  return s;
}

char* fgets(char *s, int size, FILE *fp)
{
  int err;
  int i;
  char c;

  for (i = 0; i < size-1; i++) {
    err = fread(&c, sizeof(c), 1, fp);
    if (err <= 0) {
      if (i != 0) {
        break;
      }
      return NULL;
    }
    if (c == '\n') {
      s[i] = c;
      i++;
      break;
    }
    s[i] = c;
  }
  s[i] = '\0';

  return s;
}

int ungetc(int c, FILE *fp)
{
  if (fp->buffer_start_index >= 1) {
    fp->buffer_start_index--;
    fp->buf[fp->buffer_start_index] = (unsigned char)c;
  }
  else {
    fp->current_record_pos--;
    memcpy(&(fp->buf[1]), &(fp->buf[0]), BTRON_FILEIO_BUFSIZE-1);
    fp->buf[0] = (unsigned char)c;
    if (fp->buffer_end_index <= BTRON_FILEIO_BUFSIZE-1) {
      fp->buffer_end_index++;
    }
  }
  return (unsigned char)c;
}

size_t fwrite(const B* ptr, size_t size, size_t nmemb, FILE *fp)
{
  int wsize = size * nmemb;
  int len = BTRON_FILEIO_BUFSIZE - fp->buffer_end_index;
  W err;

  if (len > 0 && wsize < len) {
    memcpy(&(fp->buf[fp->buffer_end_index]), ptr, wsize);
    fp->buffer_end_index += wsize;
    return nmemb;
  }
  else if (len > 0 && wsize >= len) {
    memcpy(&(fp->buf[fp->buffer_end_index]), ptr, len);
    wsize -= len;
    ptr += len;
    err = wri_rec(fp->fd, fp->record_size, &(fp->buf[fp->buffer_start_index]),
                  BTRON_FILEIO_BUFSIZE - fp->buffer_start_index, NULL, NULL, 0);
    if (err < 0) {
      return 0;
    }
    fp->record_size += BTRON_FILEIO_BUFSIZE - fp->buffer_start_index;
    fp->buffer_start_index = 0;
    fp->buffer_end_index = 0;

    if (wsize == 0) {
      return nmemb;
    }
  }

  assert(fp->buffer_start_index == fp->buffer_end_index);
  fp->buffer_start_index = 0;
  fp->buffer_end_index = 0;

  /* write/copy remaining wsize-bytes */
  if (wsize >= BTRON_FILEIO_BUFSIZE) {
    err = wri_rec(fp->fd, fp->record_size, ptr, wsize, NULL, NULL, 0);
    if (err < 0) {
      return 0;
    }
    fp->record_size += wsize;
  }
  else {
    memcpy(fp->buf, ptr, wsize);
    fp->buffer_end_index = wsize;
  }
  return nmemb;
}

int fputc(int c, FILE *fp)
{
  int err;
  unsigned char uc = c;

  err = fwrite(&uc, sizeof(uc), 1, fp);
  if (err <= 0) {
    return EOF;
  }
  return uc;
}

int fputs(const char *s, FILE *fp)
{
  int slen = strlen(s);
  int err;

  err = fwrite(s, sizeof(char), slen, fp);
  if (err <= 0) {
    return EOF;
  }
  return 0;
}
